@@ -1,7 +1,21 @@
 # -*- coding: utf-8 -*-
 #
 # Picard, the next-generation MusicBrainz tagger
-# Copyright (C) 2006-2007 Lukáš Lalinský
+#
+# Copyright (C) 2006-2009, 2011-2012 Lukáš Lalinský
+# Copyright (C) 2008-2011, 2014, 2018-2020 Philipp Wolfer
+# Copyright (C) 2009 Carlin Mangar
+# Copyright (C) 2011-2012 Johannes Weißl
+# Copyright (C) 2011-2014 Michael Wiencek
+# Copyright (C) 2011-2014 Wieland Hoffmann
+# Copyright (C) 2013 Calvin Walton
+# Copyright (C) 2013-2014, 2017-2019 Laurent Monin
+# Copyright (C) 2013-2015, 2017 Sophist-UK
+# Copyright (C) 2015 Frederik “Freso” S. Olesen
+# Copyright (C) 2016 Christoph Reiter
+# Copyright (C) 2016-2018 Sambhav Kothari
+# Copyright (C) 2017 tungol
+# Copyright (C) 2019 Zenara Daley
 #
 # This program is free software; you can redistribute it and/or
 # modify it under the terms of the GNU General Public License
@@ -17,6 +31,7 @@
 # along with this program; if not, write to the Free Software
 # Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 
+
 from collections import defaultdict
 import re
 from urllib.parse import urlparse
@@ -37,32 +52,39 @@ from picard.coverart.image import (
     TagCoverArtImage,
 )
 from picard.file import File
-from picard.formats.mutagenext import compatid3
+from picard.formats.mutagenext import (
+    compatid3,
+    delall_ci,
+)
 from picard.metadata import Metadata
 from picard.util import (
     encode_filename,
     sanitize_date,
 )
+from picard.util.tags import parse_comment_tag
+
 
 id3.GRP1 = compatid3.GRP1
 id3.TCMP = compatid3.TCMP
 id3.TSO2 = compatid3.TSO2
 id3.TSOC = compatid3.TSOC
 
-__ID3_IMAGE_TYPE_MAP = {
-    "other": 0,
-    "obi": 0,
-    "tray": 0,
-    "spine": 0,
-    "sticker": 0,
-    "front": 3,
-    "back": 4,
-    "booklet": 5,
-    "medium": 6,
-    "track": 6,
-}
+__IMAGE_TYPES = [
+    ("obi", 0),
+    ("tray", 0),
+    ("spine", 0),
+    ("sticker", 0),
+    ("other", 0),
+    ("front", 3),
+    ("back", 4),
+    ("booklet", 5),
+    ("track", 6),
+    ("medium", 6),
+]
 
-__ID3_REVERSE_IMAGE_TYPE_MAP = dict([(v, k) for k, v in __ID3_IMAGE_TYPE_MAP.items()])
+__ID3_IMAGE_TYPE_MAP = dict(__IMAGE_TYPES)
+
+__ID3_REVERSE_IMAGE_TYPE_MAP = dict([(v, k) for k, v in __IMAGE_TYPES])
 
 
 def id3text(text, encoding):
@@ -133,6 +155,7 @@ class ID3File(File):
         'COMM': 'comment',
         'TOAL': 'originalalbum',
         'TOPE': 'originalartist',
+        'TOFN': 'originalfilename',
 
         # The following are informal iTunes extensions to id3v2:
         'TCMP': 'compilation',
@@ -174,6 +197,18 @@ class ID3File(File):
     __rtranslate_freetext = dict([(v, k) for k, v in __translate_freetext.items()])
     __translate_freetext['writer'] = 'writer'  # For backward compatibility of case
 
+    # Freetext fields that are loaded case-insensitive
+    __rtranslate_freetext_ci = {
+        'replaygain_album_gain': 'REPLAYGAIN_ALBUM_GAIN',
+        'replaygain_album_peak': 'REPLAYGAIN_ALBUM_PEAK',
+        'replaygain_album_range': 'REPLAYGAIN_ALBUM_RANGE',
+        'replaygain_track_gain': 'REPLAYGAIN_TRACK_GAIN',
+        'replaygain_track_peak': 'REPLAYGAIN_TRACK_PEAK',
+        'replaygain_track_range': 'REPLAYGAIN_TRACK_RANGE',
+        'replaygain_reference_loudness': 'REPLAYGAIN_REFERENCE_LOUDNESS',
+    }
+    __translate_freetext_ci = dict([(b.lower(), a) for a, b in __rtranslate_freetext_ci.items()])
+
     # Obsolete tag names which will still be loaded, but will get renamed on saving
     __rename_freetext = {
         'Artists': 'ARTISTS',
@@ -199,6 +234,10 @@ class ID3File(File):
         'MVIN': re.compile(r'^(?P<movementnumber>\d+)(?:/(?P<movementtotal>\d+))?$')
     }
 
+    def __init__(self, filename):
+        super().__init__(filename)
+        self.__casemap = {}
+
     def build_TXXX(self, encoding, desc, values):
         """Construct and return a TXXX frame."""
         # This is here so that plugins can customize the behavior of TXXX
@@ -211,6 +250,7 @@ class ID3File(File):
 
     def _load(self, filename):
         log.debug("Loading file %r", filename)
+        self.__casemap = {}
         file = self._get_file(encode_filename(filename))
         tags = file.tags or {}
         # upgrade custom 2.3 frames to 2.4
@@ -226,19 +266,23 @@ class ID3File(File):
                 if frameid.startswith('T') or frameid in ["GRP1", "MVNM"]:
                     for text in frame.text:
                         if text:
-                            metadata.add(name, str(text))
+                            metadata.add(name, text)
                 elif frameid == 'COMM':
                     for text in frame.text:
                         if text:
-                            metadata.add('%s:%s' % (name, frame.desc), str(text))
+                            if frame.lang == 'eng':
+                                name = '%s:%s' % (name, frame.desc)
+                            else:
+                                name = '%s:%s:%s' % (name, frame.lang, frame.desc)
+                            metadata.add(name, text)
                 else:
-                    metadata.add(name, str(frame))
+                    metadata.add(name, frame)
             elif frameid == 'TIT1':
                 itunes_compatible = config.setting['itunes_compatible_grouping']
                 name = 'work' if itunes_compatible else 'grouping'
                 for text in frame.text:
                     if text:
-                        metadata.add(name, str(text))
+                        metadata.add(name, text)
             elif frameid == "TMCL":
                 for role, name in frame.people:
                     if role or name:
@@ -253,12 +297,17 @@ class ID3File(File):
                         metadata.add('performer:%s' % role, name)
             elif frameid == 'TXXX':
                 name = frame.desc
+                name_lower = name.lower()
                 if name in self.__rename_freetext:
                     name = self.__rename_freetext[name]
-                if name in self.__translate_freetext:
+                if name_lower in self.__translate_freetext_ci:
+                    orig_name = name
+                    name = self.__translate_freetext_ci[name_lower]
+                    self.__casemap[name] = orig_name
+                elif name in self.__translate_freetext:
                     name = self.__translate_freetext[name]
-                elif ((name in self.__rtranslate) !=
-                        (name in self.__rtranslate_freetext)):
+                elif ((name in self.__rtranslate)
+                      != (name in self.__rtranslate_freetext)):
                     # If the desc of a TXXX frame conflicts with the name of a
                     # Picard tag, load it into ~id3:TXXX:desc rather than desc.
                     #
@@ -268,12 +317,12 @@ class ID3File(File):
                     # ways.) Currently, the only tag in both is license.
                     name = '~id3:TXXX:' + name
                 for text in frame.text:
-                    metadata.add(name, str(text))
+                    metadata.add(name, text)
             elif frameid == 'USLT':
                 name = 'lyrics'
                 if frame.desc:
                     name += ':%s' % frame.desc
-                metadata.add(name, str(frame.text))
+                metadata.add(name, frame.text)
             elif frameid == 'UFID' and frame.owner == 'http://musicbrainz.org':
                 metadata['musicbrainz_recordingid'] = frame.data.decode('ascii', 'ignore')
             elif frameid in self.__tag_re_parse.keys():
@@ -297,11 +346,11 @@ class ID3File(File):
                 except CoverArtImageError as e:
                     log.error('Cannot load image from %r: %s' % (filename, e))
                 else:
-                    metadata.append_image(coverartimage)
+                    metadata.images.append(coverartimage)
             elif frameid == 'POPM':
                 # Rating in ID3 ranges from 0 to 255, normalize this to the range 0 to 5
                 if frame.email == config.setting['rating_user_email']:
-                    rating = str(int(round(frame.rating / 255.0 * (config.setting['rating_steps'] - 1))))
+                    rating = int(round(frame.rating / 255.0 * (config.setting['rating_steps'] - 1)))
                     metadata.add('~rating', rating)
 
         if 'date' in metadata:
@@ -319,7 +368,8 @@ class ID3File(File):
 
         if config.setting['clear_existing_tags']:
             tags.clear()
-        if metadata.images_to_be_saved_to_tags:
+        images_to_save = list(metadata.images.to_be_saved_to_tags())
+        if images_to_save:
             tags.delall('APIC')
 
         encoding = {'utf-8': 3, 'utf-16': 1}.get(config.setting['id3v2_encoding'], 0)
@@ -350,7 +400,7 @@ class ID3File(File):
         # impossible to save two images, even of different types, without
         # any description.
         counters = defaultdict(lambda: 0)
-        for image in metadata.images_to_be_saved_to_tags:
+        for image in images_to_save:
             desc = desctag = image.comment
             if counters[desc] > 0:
                 if desc:
@@ -366,7 +416,6 @@ class ID3File(File):
 
         tmcl = mutagen.id3.TMCL(encoding=encoding, people=[])
         tipl = mutagen.id3.TIPL(encoding=encoding, people=[])
-
         for name, values in metadata.rawitems():
             values = [id3text(v, encoding) for v in values]
             name = id3text(name, encoding)
@@ -381,13 +430,13 @@ class ID3File(File):
                         tipl.people.append([role, value])
                     else:
                         tmcl.people.append([role, value])
-            elif name.startswith('comment:'):
-                desc = name.split(':', 1)[1]
+            elif name == 'comment' or name.startswith('comment:'):
+                (lang, desc) = parse_comment_tag(name)
                 if desc.lower()[:4] == 'itun':
                     tags.delall('COMM:' + desc)
                     tags.add(id3.COMM(encoding=0, desc=desc, lang='eng', text=[v + '\x00' for v in values]))
                 else:
-                    tags.add(id3.COMM(encoding=encoding, desc=desc, lang='eng', text=values))
+                    tags.add(id3.COMM(encoding=encoding, desc=desc, lang=lang, text=values))
             elif name.startswith('lyrics:') or name == 'lyrics':
                 if ':' in name:
                     desc = name.split(':', 1)[1]
@@ -447,6 +496,13 @@ class ID3File(File):
                         tags.delall('XSOP')
                     elif frameid == 'TSO2':
                         tags.delall('TXXX:ALBUMARTISTSORT')
+            elif name in self.__rtranslate_freetext_ci:
+                if name in self.__casemap:
+                    description = self.__casemap[name]
+                else:
+                    description = self.__rtranslate_freetext_ci[name]
+                delall_ci(tags, 'TXXX:' + description)
+                tags.add(self.build_TXXX(encoding, description, values))
             elif name in self.__rtranslate_freetext:
                 description = self.__rtranslate_freetext[name]
                 if description in self.__rrename_freetext:
@@ -492,10 +548,11 @@ class ID3File(File):
                                 if people[0] == role:
                                     frame.people.remove(people)
                 elif name.startswith('comment:'):
-                    desc = name.split(':', 1)[1]
+                    (lang, desc) = parse_comment_tag(name)
                     if desc.lower()[:4] != 'itun':
                         for key, frame in list(tags.items()):
-                            if frame.FrameID == 'COMM' and frame.desc == desc:
+                            if (frame.FrameID == 'COMM' and frame.desc == desc
+                                and frame.lang == lang):
                                 del tags[key]
                 elif name.startswith('lyrics:') or name == 'lyrics':
                     if ':' in name:
@@ -503,7 +560,7 @@ class ID3File(File):
                     else:
                         desc = ''
                     for key, frame in list(tags.items()):
-                        if frame.FrameID == desc:
+                        if frame.FrameID == 'USLT' and frame.desc == desc:
                             del tags[key]
                 elif name in self._rtipl_roles:
                     role = self._rtipl_roles[name]
@@ -538,7 +595,7 @@ class ID3File(File):
 
     @classmethod
     def supports_tag(cls, name):
-        unsupported_tags = {}
+        unsupported_tags = ['r128_album_gain', 'r128_track_gain']
         return ((name and not name.startswith("~") and name not in unsupported_tags)
                 or name in ("~rating", "~length")
                 or name.startswith("~id3"))
