@@ -1,8 +1,30 @@
 # -*- coding: utf-8 -*-
 #
 # Picard, the next-generation MusicBrainz tagger
+#
 # Copyright (C) 2004 Robert Kaye
-# Copyright (C) 2006 Lukáš Lalinský
+# Copyright (C) 2006-2009, 2011-2014, 2017 Lukáš Lalinský
+# Copyright (C) 2008 Gary van der Merwe
+# Copyright (C) 2008 amckinle
+# Copyright (C) 2008-2010, 2014-2015, 2018-2020 Philipp Wolfer
+# Copyright (C) 2009 Carlin Mangar
+# Copyright (C) 2010 Andrew Barnert
+# Copyright (C) 2011-2014 Michael Wiencek
+# Copyright (C) 2011-2014, 2017-2019 Wieland Hoffmann
+# Copyright (C) 2012 Chad Wilson
+# Copyright (C) 2013 Calvin Walton
+# Copyright (C) 2013 Ionuț Ciocîrlan
+# Copyright (C) 2013 brainz34
+# Copyright (C) 2013-2014, 2017 Sophist-UK
+# Copyright (C) 2013-2015, 2017-2019 Laurent Monin
+# Copyright (C) 2016 Rahul Raturi
+# Copyright (C) 2016 Simon Legner
+# Copyright (C) 2016 Suhas
+# Copyright (C) 2016-2018 Sambhav Kothari
+# Copyright (C) 2017-2018 Vishal Choudhary
+# Copyright (C) 2018 Bob Swift
+# Copyright (C) 2018 virusMac
+# Copyright (C) 2019 Joel Lintunen
 #
 # This program is free software; you can redistribute it and/or
 # modify it under the terms of the GNU General Public License
@@ -18,6 +40,7 @@
 # along with this program; if not, write to the Free Software
 # Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 
+
 import argparse
 from functools import partial
 from itertools import chain
@@ -35,9 +58,12 @@ from PyQt5 import (
     QtGui,
     QtWidgets,
 )
+from PyQt5.QtDBus import QDBusConnection
 
 from picard import (
+    PICARD_APP_ID,
     PICARD_APP_NAME,
+    PICARD_DESKTOP_NAME,
     PICARD_FANCY_VERSION_STR,
     PICARD_ORG_NAME,
     acoustid,
@@ -48,6 +74,7 @@ from picard.acoustid.manager import AcoustIDManager
 from picard.album import (
     Album,
     NatAlbum,
+    run_album_post_removal_processors,
 )
 from picard.browser.browser import BrowserIntegration
 from picard.browser.filelookup import FileLookup
@@ -62,12 +89,18 @@ from picard.const import (
     USER_DIR,
     USER_PLUGIN_DIR,
 )
+from picard.const.sys import (
+    IS_FROZEN,
+    IS_HAIKU,
+    IS_MACOS,
+    IS_WIN,
+)
 from picard.dataobj import DataObject
 from picard.disc import Disc
 from picard.file import File
 from picard.formats import open_ as open_file
 from picard.i18n import setup_gettext
-from picard.plugin import PluginManager
+from picard.pluginmanager import PluginManager
 from picard.releasegroup import ReleaseGroup
 from picard.track import (
     NonAlbumTrack,
@@ -91,7 +124,7 @@ from picard.webservice.api_helpers import (
     MBAPIHelper,
 )
 
-import picard.resources  # pylint: disable=unused-import
+import picard.resources  # noqa: F401 # pylint: disable=unused-import
 
 from picard.ui.itemviews import BaseTreeView
 from picard.ui.mainwindow import MainWindow
@@ -130,18 +163,13 @@ class Tagger(QtWidgets.QApplication):
 
         # Use the new fusion style from PyQt5 for a modern and consistent look
         # across all OSes.
-        if sys.platform != "darwin":
+        if not IS_MACOS and not IS_HAIKU:
             self.setStyle('Fusion')
 
-        # Set the WM_CLASS to 'MusicBrainz-Picard' so desktop environments
-        # can use it to look up the app
-        super().__init__(['MusicBrainz-Picard'] + unparsed_args)
+        super().__init__(sys.argv)
         self.__class__.__instance = self
         config._setup(self, picard_args.config_file)
 
-        # Allow High DPI Support
-        self.setAttribute(QtCore.Qt.AA_UseHighDpiPixmaps)
-        self.setAttribute(QtCore.Qt.AA_EnableHighDpiScaling)
         super().setStyleSheet(
             'QGroupBox::title { /* PICARD-1206, Qt bug workaround */ }'
         )
@@ -171,7 +199,7 @@ class Tagger(QtWidgets.QApplication):
         self.save_thread_pool = QtCore.QThreadPool(self)
         self.save_thread_pool.setMaxThreadCount(1)
 
-        if not sys.platform == "win32":
+        if not IS_WIN:
             # Set up signal handling
             # It's not possible to call all available functions from signal
             # handlers, therefore we need to set up a QSocketNotifier to listen
@@ -194,9 +222,11 @@ class Tagger(QtWidgets.QApplication):
             signal.signal(signal.SIGINT, self.signal)
             signal.signal(signal.SIGTERM, self.signal)
 
-        if sys.platform == "darwin":
+        if IS_MACOS:
             # On macOS it is not common that the global menu shows icons
             self.setAttribute(QtCore.Qt.AA_DontShowIconsInMenus)
+            # Ensure monospace font works on macOS
+            QtGui.QFont.insertSubstitution('Monospace', 'Menlo')
 
         # Setup logging
         log.debug("Starting Picard from %r", os.path.abspath(__file__))
@@ -218,32 +248,32 @@ class Tagger(QtWidgets.QApplication):
         # translated
         setup_gettext(localedir, config.setting["ui_language"], log.debug)
 
-        upgrade_config()
+        upgrade_config(config.config)
 
         self.webservice = WebService()
         self.mb_api = MBAPIHelper(self.webservice)
-        self.acoustid_api = AcoustIdAPIHelper(self.webservice)
 
         load_user_collections()
 
         # Initialize fingerprinting
-        self._acoustid = acoustid.AcoustIDClient()
+        acoustid_api = AcoustIdAPIHelper(self.webservice)
+        self._acoustid = acoustid.AcoustIDClient(acoustid_api)
         self._acoustid.init()
+        self.acoustidmanager = AcoustIDManager(acoustid_api)
 
         # Load plugins
         self.pluginmanager = PluginManager()
         if not self._no_plugins:
-            if hasattr(sys, "frozen"):
-                self.pluginmanager.load_plugindir(os.path.join(os.path.dirname(sys.argv[0]), "plugins"))
+            if IS_FROZEN:
+                self.pluginmanager.load_plugins_from_directory(os.path.join(os.path.dirname(sys.argv[0]), "plugins"))
             else:
                 mydir = os.path.dirname(os.path.abspath(__file__))
-                self.pluginmanager.load_plugindir(os.path.join(mydir, "plugins"))
+                self.pluginmanager.load_plugins_from_directory(os.path.join(mydir, "plugins"))
 
             if not os.path.exists(USER_PLUGIN_DIR):
                 os.makedirs(USER_PLUGIN_DIR)
-            self.pluginmanager.load_plugindir(USER_PLUGIN_DIR)
+            self.pluginmanager.load_plugins_from_directory(USER_PLUGIN_DIR)
 
-        self.acoustidmanager = AcoustIDManager()
         self.browser_integration = BrowserIntegration()
 
         self.files = {}
@@ -278,16 +308,20 @@ class Tagger(QtWidgets.QApplication):
         webbrowser2.open(authorization_url)
         if not parent:
             parent = self.window
-        authorization_code, ok = QtWidgets.QInputDialog.getText(parent,
-            _("MusicBrainz Account"), _("Authorization code:"))
-        if ok:
+        dialog = QtWidgets.QInputDialog(parent)
+        dialog.setWindowModality(QtCore.Qt.WindowModal)
+        dialog.setWindowTitle(_("MusicBrainz Account"))
+        dialog.setLabelText(_("Authorization code:"))
+        status = dialog.exec_()
+        if status == QtWidgets.QDialog.Accepted:
+            authorization_code = dialog.textValue()
             self.webservice.oauth_manager.exchange_authorization_code(
                 authorization_code, scopes,
                 partial(self.on_mb_authorization_finished, callback))
         else:
             callback(False)
 
-    def on_mb_authorization_finished(self, callback, successful):
+    def on_mb_authorization_finished(self, callback, successful=False):
         if successful:
             self.webservice.oauth_manager.fetch_username(
                 partial(self.on_mb_login_finished, callback))
@@ -322,7 +356,8 @@ class Tagger(QtWidgets.QApplication):
         """Move `file` to recording `recordingid` on album `albumid`."""
         album = self.load_album(albumid)
         file.move(album.unmatched_files)
-        album.run_when_loaded(partial(album.match_file, file, recordingid))
+        album.run_when_loaded(partial(album.match_files, [file],
+                                      recordingid=recordingid))
 
     def create_nats(self):
         if self.nats is None:
@@ -379,7 +414,11 @@ class Tagger(QtWidgets.QApplication):
         if isinstance(event, thread.ProxyToMainEvent):
             event.run()
         elif event.type() == QtCore.QEvent.FileOpen:
-            self.add_files([event.file()])
+            file = event.file()
+            if os.path.isdir(file):
+                self.add_directory(file)
+            else:
+                self.add_files([file])
             # We should just return True here, except that seems to
             # cause the event's sender to get a -9874 error, so
             # apparently there's some magic inside QFileOpenEvent...
@@ -387,25 +426,42 @@ class Tagger(QtWidgets.QApplication):
         return super().event(event)
 
     def _file_loaded(self, file, target=None):
-        if file is not None and not file.has_error():
-            recordingid = file.metadata.getall('musicbrainz_recordingid')[0] \
-                if 'musicbrainz_recordingid' in file.metadata else ''
-            if target is not None:
-                self.move_files([file], target)
-            elif not config.setting["ignore_file_mbids"]:
-                albumid = file.metadata.getall('musicbrainz_albumid')[0] \
-                    if 'musicbrainz_albumid' in file.metadata else ''
-                if mbid_validate(albumid):
-                    if mbid_validate(recordingid):
-                        self.move_file_to_track(file, albumid, recordingid)
-                    else:
-                        self.move_file_to_album(file, albumid)
-                elif mbid_validate(recordingid):
-                    self.move_file_to_nat(file, recordingid)
-                elif config.setting['analyze_new_files'] and file.can_analyze():
-                    self.analyze([file])
-            elif config.setting['analyze_new_files'] and file.can_analyze():
-                self.analyze([file])
+        if file is None or file.has_error():
+            return
+
+        if target is not None:
+            self.move_files([file], target)
+            return
+
+        if not config.setting["ignore_file_mbids"]:
+            recordingid = file.metadata['musicbrainz_recordingid']
+            is_valid_recordingid = mbid_validate(recordingid)
+
+            albumid = file.metadata['musicbrainz_albumid']
+            is_valid_albumid = mbid_validate(albumid)
+
+            if is_valid_albumid and is_valid_recordingid:
+                log.debug("%r has release (%s) and recording (%s) MBIDs, moving to track...",
+                          file, albumid, recordingid)
+                self.move_file_to_track(file, albumid, recordingid)
+                return
+
+            if is_valid_albumid:
+                log.debug("%r has only release MBID (%s), moving to album...",
+                          file, albumid)
+                self.move_file_to_album(file, albumid)
+                return
+
+            if is_valid_recordingid:
+                log.debug("%r has only recording MBID (%s), moving to non-album track...",
+                          file, recordingid)
+                self.move_file_to_nat(file, recordingid)
+                return
+
+        # fallback on analyze if nothing else worked
+        if config.setting['analyze_new_files'] and file.can_analyze():
+            log.debug("Trying to analyze %r ...", file)
+            self.analyze([file])
 
     def move_files(self, files, target):
         if target is None:
@@ -414,9 +470,11 @@ class Tagger(QtWidgets.QApplication):
         if isinstance(target, (Track, Cluster)):
             for file in files:
                 file.move(target)
+                QtCore.QCoreApplication.processEvents()
         elif isinstance(target, File):
             for file in files:
                 file.move(target.parent)
+                QtCore.QCoreApplication.processEvents()
         elif isinstance(target, Album):
             self.move_files_to_album(files, album=target)
         elif isinstance(target, ClusterList):
@@ -427,7 +485,10 @@ class Tagger(QtWidgets.QApplication):
         ignoreregex = None
         pattern = config.setting['ignore_regex']
         if pattern:
-            ignoreregex = re.compile(pattern)
+            try:
+                ignoreregex = re.compile(pattern)
+            except re.error as e:
+                log.error("Failed evaluating regular expression for ignore_regex: %s", e)
         ignore_hidden = config.setting["ignore_hidden_files"]
         new_files = []
         for filename in filenames:
@@ -652,10 +713,13 @@ class Tagger(QtWidgets.QApplication):
                 self._acoustid.stop_analyze(file)
                 del self.files[file.filename]
                 file.remove(from_parent)
+        self.tagger_stats_changed.emit()
 
     def remove_album(self, album):
         """Remove the specified album."""
         log.debug("Removing %r", album)
+        if album.id not in self.albums:
+            return
         album.stop_loading()
         self.remove_files(self.get_files_from_objects([album]))
         del self.albums[album.id]
@@ -664,6 +728,8 @@ class Tagger(QtWidgets.QApplication):
         if album == self.nats:
             self.nats = None
         self.album_removed.emit(album)
+        run_album_post_removal_processors(album)
+        self.tagger_stats_changed.emit()
 
     def remove_nat(self, track):
         """Remove the specified non-album track."""
@@ -724,7 +790,7 @@ class Tagger(QtWidgets.QApplication):
     def lookup_cd(self, action):
         """Reads CD from the selected drive and tries to lookup the DiscID on MusicBrainz."""
         if isinstance(action, QtWidgets.QAction):
-            device = action.text()
+            device = action.data()
         elif config.setting["cd_lookup_device"] != '':
             device = config.setting["cd_lookup_device"].split(",", 1)[0]
         else:
@@ -744,11 +810,26 @@ class Tagger(QtWidgets.QApplication):
 
     def analyze(self, objs):
         """Analyze the file(s)."""
+        if not self.use_acoustid:
+            return
         files = self.get_files_from_objects(objs)
         for file in files:
             file.set_pending()
-            if self.use_acoustid:
-                self._acoustid.analyze(file, partial(file._lookup_finished, 'acoustid'))
+            self._acoustid.analyze(file, partial(file._lookup_finished,
+                                                 File.LOOKUP_ACOUSTID))
+
+    def generate_fingerprints(self, objs):
+        """Generate the fingerprints without matching the files."""
+        if not self.use_acoustid:
+            return
+        files = self.get_files_from_objects(objs)
+
+        def finished(file, result):
+            file.clear_pending()
+
+        for file in files:
+            file.set_pending()
+            self._acoustid.fingerprint(file, partial(finished, file))
 
     # =======================================================================
     #  Metadata-based lookups
@@ -775,6 +856,7 @@ class Tagger(QtWidgets.QApplication):
             cluster = self.load_cluster(name, artist)
             for file in sorted(files, key=attrgetter('discnumber', 'tracknumber', 'base_filename')):
                 file.move(cluster)
+                QtCore.QCoreApplication.processEvents()
 
     def load_cluster(self, name, artist):
         for cluster in self.clusters:
@@ -838,6 +920,14 @@ def process_picard_args():
     parser = argparse.ArgumentParser(
         epilog="If one of the filenames begins with a hyphen, use -- to separate the options from the filenames."
     )
+    # Qt default arguments. Parse them so Picard does not interpret the
+    # arguments as file names to load.
+    parser.add_argument("-style", nargs=1, help=argparse.SUPPRESS)
+    parser.add_argument("-stylesheet", nargs=1, help=argparse.SUPPRESS)
+    # Same for default X arguments
+    parser.add_argument("-display", nargs=1, help=argparse.SUPPRESS)
+
+    # Picard specific arguments
     parser.add_argument("-c", "--config-file", action='store',
                         default=None,
                         help="location of the configuration file")
@@ -860,6 +950,14 @@ def main(localedir=None, autoupdate=True):
     # Some libs (ie. Phonon) require those to be set
     QtWidgets.QApplication.setApplicationName(PICARD_APP_NAME)
     QtWidgets.QApplication.setOrganizationName(PICARD_ORG_NAME)
+    QtWidgets.QApplication.setDesktopFileName(PICARD_DESKTOP_NAME)
+
+    # Allow High DPI Support
+    QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_UseHighDpiPixmaps)
+    QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_EnableHighDpiScaling)
+
+    # Enable mnemonics on all platforms, even macOS
+    QtGui.qt_set_sequence_auto_mnemonic(True)
 
     signal.signal(signal.SIGINT, signal.SIG_DFL)
 
@@ -869,6 +967,10 @@ def main(localedir=None, autoupdate=True):
     if picard_args.long_version:
         return longversion()
 
+    if not (IS_WIN or IS_MACOS or IS_HAIKU):
+        dbus = QDBusConnection.sessionBus()
+        dbus.registerService(PICARD_APP_ID)
+
     tagger = Tagger(picard_args, unparsed_args, localedir, autoupdate)
 
     # Initialize Qt default translations
