@@ -1,8 +1,18 @@
 # -*- coding: utf-8 -*-
 #
 # Picard, the next-generation MusicBrainz tagger
+#
 # Copyright (C) 2004 Robert Kaye
-# Copyright (C) 2006 Lukáš Lalinský
+# Copyright (C) 2006-2007, 2012 Lukáš Lalinský
+# Copyright (C) 2011-2014 Michael Wiencek
+# Copyright (C) 2012 Nikolai Prokoschenko
+# Copyright (C) 2013-2014 Sophist-UK
+# Copyright (C) 2013-2014, 2017-2019 Laurent Monin
+# Copyright (C) 2015 Ohm Patel
+# Copyright (C) 2015 Wieland Hoffmann
+# Copyright (C) 2015, 2018-2020 Philipp Wolfer
+# Copyright (C) 2016-2018 Sambhav Kothari
+# Copyright (C) 2018 Vishal Choudhary
 #
 # This program is free software; you can redistribute it and/or
 # modify it under the terms of the GNU General Public License
@@ -18,6 +28,7 @@
 # along with this program; if not, write to the Free Software
 # Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 
+
 from collections import defaultdict
 from functools import partial
 
@@ -41,29 +52,24 @@ from picard.util import (
     thread,
     throttle,
 )
+from picard.util.preservedtags import PreservedTags
 from picard.util.tags import display_tag_name
 
-from picard.ui.edittagdialog import EditTagDialog
-
-COMMON_TAGS = [
-    "title",
-    "artist",
-    "album",
-    "tracknumber",
-    "~length",
-    "date",
-]
+from picard.ui.colors import interface_colors
+from picard.ui.edittagdialog import (
+    EditTagDialog,
+    TagEditorDelegate,
+)
 
 
 class TagStatus:
 
-    NoChange = 1
-    Added = 2
-    Removed = 4
-    # Added | Removed = Changed
-    Changed = 6
-    Empty = 8
-    NotRemovable = 16
+    NOCHANGE = 1
+    ADDED = 2
+    REMOVED = 4
+    CHANGED = ADDED | REMOVED
+    EMPTY = 8
+    NOTREMOVABLE = 16
 
 
 class TagCounter(dict):
@@ -131,53 +137,51 @@ class TagDiff(object):
             self.new.add(tag, new_values)
 
         if (orig_values and not new_values) or removed:
-            self.status[tag] |= TagStatus.Removed
+            self.status[tag] |= TagStatus.REMOVED
         elif new_values and not orig_values:
-            self.status[tag] |= TagStatus.Added
+            self.status[tag] |= TagStatus.ADDED
             removable = True
         elif orig_values and new_values and self.__tag_ne(tag, orig_values, new_values):
-            self.status[tag] |= TagStatus.Changed
-        elif not (orig_values or new_values or tag in COMMON_TAGS):
-            self.status[tag] |= TagStatus.Empty
+            self.status[tag] |= TagStatus.CHANGED
+        elif not (orig_values or new_values or tag in config.setting['metadatabox_top_tags']):
+            self.status[tag] |= TagStatus.EMPTY
         else:
-            self.status[tag] |= TagStatus.NoChange
+            self.status[tag] |= TagStatus.NOCHANGE
 
         if not removable:
-            self.status[tag] |= TagStatus.NotRemovable
+            self.status[tag] |= TagStatus.NOTREMOVABLE
 
     def tag_status(self, tag):
         status = self.status[tag]
-        for s in (TagStatus.Changed, TagStatus.Added,
-                  TagStatus.Removed, TagStatus.Empty):
+        for s in (TagStatus.CHANGED, TagStatus.ADDED,
+                  TagStatus.REMOVED, TagStatus.EMPTY):
             if status & s == s:
                 return s
-        return TagStatus.NoChange
+        return TagStatus.NOCHANGE
 
 
-class PreservedTags:
+class TableTagEditorDelegate(TagEditorDelegate):
 
-    opt_name = 'preserved_tags'
+    def createEditor(self, parent, option, index):
+        editor = super().createEditor(parent, option, index)
+        if editor and isinstance(editor, QtWidgets.QPlainTextEdit):
+            table = self.parent()
+            # Set the editor to the row height, but at least 80 pixel
+            # to allow for proper multiline editing.
+            height = max(80, table.rowHeight(index.row()) - 1)
+            editor.setMinimumSize(QtCore.QSize(0, height))
+            # Resize the row so the editor fits in. Add 1 pixel, otherwise the
+            # frame gets hidden.
+            table.setRowHeight(index.row(), editor.frameSize().height() + 1)
+        return editor
 
-    def __init__(self):
-        self._tags = self._from_config()
+    def sizeHint(self, option, index):
+        # Expand the row for multiline content, but limit the maximum row height.
+        size_hint = super().sizeHint(option, index)
+        return QtCore.QSize(size_hint.width(), min(160, size_hint.height()))
 
-    def _to_config(self):
-        config.setting[self.opt_name] = ", ".join(sorted(self._tags))
-
-    def _from_config(self):
-        tags = config.setting[self.opt_name].split(',')
-        return set(filter(bool, map(str.strip, tags)))
-
-    def add(self, name):
-        self._tags.add(name)
-        self._to_config()
-
-    def discard(self, name):
-        self._tags.discard(name)
-        self._to_config()
-
-    def __contains__(self, key):
-        return key in self._tags
+    def get_tag_name(self, index):
+        return index.data(QtCore.Qt.UserRole)
 
 
 class MetadataBox(QtWidgets.QTableWidget):
@@ -204,12 +208,8 @@ class MetadataBox(QtWidgets.QTableWidget):
         self.setTabKeyNavigation(False)
         self.setStyleSheet("QTableWidget {border: none;}")
         self.setAttribute(QtCore.Qt.WA_MacShowFocusRect, 1)
-        self.colors = {
-            TagStatus.NoChange: self.palette().color(QtGui.QPalette.Text),
-            TagStatus.Removed: QtGui.QBrush(QtGui.QColor("red")),
-            TagStatus.Added: QtGui.QBrush(QtGui.QColor("green")),
-            TagStatus.Changed: QtGui.QBrush(QtGui.QColor("darkgoldenrod"))
-        }
+        self.setItemDelegate(TableTagEditorDelegate(self))
+        self.setWordWrap(False)
         self.files = set()
         self.tracks = set()
         self.objects = set()
@@ -296,7 +296,7 @@ class MetadataBox(QtWidgets.QTableWidget):
         super().closeEditor(editor, hint)
         tag = self.tag_diff.tag_names[self.editing.row()]
         old = self.tag_diff.new[tag]
-        new = [editor.text()]
+        new = [self._get_editor_value(editor)]
         if old == new:
             self.editing.setText(old[0])
         else:
@@ -304,6 +304,14 @@ class MetadataBox(QtWidgets.QTableWidget):
         self.editing = None
         self.update()
 
+    @staticmethod
+    def _get_editor_value(editor):
+        if hasattr(editor, 'text'):
+            return editor.text()
+        elif hasattr(editor, 'toPlainText'):
+            return editor.toPlainText()
+        return ''
+
     def contextMenuEvent(self, event):
         menu = QtWidgets.QMenu(self)
         if self.objects:
@@ -339,8 +347,8 @@ class MetadataBox(QtWidgets.QTableWidget):
                             menu.addAction(lookup_action)
                     if self.tag_is_removable(tag):
                         removals.append(partial(self.remove_tag, tag))
-                    status = self.tag_diff.status[tag] & TagStatus.Changed
-                    if status == TagStatus.Changed or status == TagStatus.Removed:
+                    status = self.tag_diff.status[tag] & TagStatus.CHANGED
+                    if status == TagStatus.CHANGED or status == TagStatus.REMOVED:
                         for file in self.files:
                             objects = [file]
                             if file.parent in self.tracks and len(self.files & set(file.parent.linked_files)) == 1:
@@ -386,7 +394,7 @@ class MetadataBox(QtWidgets.QTableWidget):
             values = []
         if not values and self.tag_is_removable(tag):
             for obj in objects:
-                obj.metadata.delete(tag)
+                del obj.metadata[tag]
         elif values:
             for obj in objects:
                 obj.metadata[tag] = values
@@ -403,7 +411,7 @@ class MetadataBox(QtWidgets.QTableWidget):
                 self.remove_tag(tag)
 
     def tag_is_removable(self, tag):
-        return self.tag_diff.status[tag] & TagStatus.NotRemovable == 0
+        return self.tag_diff.status[tag] & TagStatus.NOTREMOVABLE == 0
 
     def selected_tags(self, discard=None):
         if discard is None:
@@ -458,6 +466,13 @@ class MetadataBox(QtWidgets.QTableWidget):
         if not (files or tracks):
             return None
 
+        self.colors = {
+            TagStatus.NOCHANGE: self.palette().color(QtGui.QPalette.Text),
+            TagStatus.REMOVED: QtGui.QBrush(interface_colors.get_qcolor('tagstatus_removed')),
+            TagStatus.ADDED: QtGui.QBrush(interface_colors.get_qcolor('tagstatus_added')),
+            TagStatus.CHANGED: QtGui.QBrush(interface_colors.get_qcolor('tagstatus_changed'))
+        }
+
         tag_diff = TagDiff(max_length_diff=config.setting["ignore_track_duration_difference_under"])
         orig_tags = tag_diff.orig
         new_tags = tag_diff.new
@@ -499,9 +514,9 @@ class MetadataBox(QtWidgets.QTableWidget):
                 tag_diff.objects += 1
 
         all_tags = set(list(orig_tags.keys()) + list(new_tags.keys()))
-        tag_names = COMMON_TAGS + \
-            sorted(all_tags.difference(COMMON_TAGS),
-                   key=lambda x: display_tag_name(x).lower())
+        common_tags = [tag for tag in config.setting['metadatabox_top_tags'] if tag in all_tags]
+        tag_names = common_tags + sorted(all_tags.difference(common_tags),
+                                         key=lambda x: display_tag_name(x).lower())
 
         if config.persist["show_changes_first"]:
             tags_by_status = {}
@@ -509,13 +524,13 @@ class MetadataBox(QtWidgets.QTableWidget):
             for tag in tag_names:
                 tags_by_status.setdefault(tag_diff.tag_status(tag), []).append(tag)
 
-            for status in (TagStatus.Changed, TagStatus.Added,
-                           TagStatus.Removed, TagStatus.NoChange):
+            for status in (TagStatus.CHANGED, TagStatus.ADDED,
+                           TagStatus.REMOVED, TagStatus.NOCHANGE):
                 tag_diff.tag_names += tags_by_status.pop(status, [])
         else:
             tag_diff.tag_names = [
                 tag for tag in tag_names if
-                tag_diff.status[tag] != TagStatus.Empty]
+                tag_diff.status[tag] != TagStatus.EMPTY]
 
         return tag_diff
 
@@ -564,7 +579,7 @@ class MetadataBox(QtWidgets.QTableWidget):
             self.set_item_value(new_item, self.tag_diff.new, name)
 
             font = new_item.font()
-            if result.tag_status(name) == TagStatus.Removed:
+            if result.tag_status(name) == TagStatus.REMOVED:
                 font.setStrikeOut(True)
             else:
                 font.setStrikeOut(False)
@@ -572,12 +587,21 @@ class MetadataBox(QtWidgets.QTableWidget):
             new_item.setFont(font)
 
             color = self.colors.get(result.tag_status(name),
-                                    self.colors[TagStatus.NoChange])
+                                    self.colors[TagStatus.NOCHANGE])
             orig_item.setForeground(color)
             new_item.setForeground(color)
 
+            alignment = QtCore.Qt.AlignLeft | QtCore.Qt.AlignTop
+            tag_item.setTextAlignment(alignment)
+            orig_item.setTextAlignment(alignment)
+            new_item.setTextAlignment(alignment)
+
+            # Adjust row height to content size
+            self.setRowHeight(i, self.sizeHintForRow(i))
+
     def set_item_value(self, item, tags, name):
         text, italic = tags.display_value(name)
+        item.setData(QtCore.Qt.UserRole, name)
         item.setText(text)
         font = item.font()
         font.setItalic(italic)
